"""mse_lib_sgx.certificate module."""

import hashlib
import ipaddress
from datetime import datetime
from pathlib import Path
from typing import List, Optional, Union, cast
from urllib.parse import urlparse

from cryptography import x509
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.serialization import (
    Encoding,
    NoEncryption,
    PrivateFormat,
    PublicFormat,
    load_pem_private_key,
)
from intel_sgx_ra.quote import Quote
from intel_sgx_ra.ratls import SGX_QUOTE_EXTENSION_OID, get_quote_from_cert

from mse_lib_sgx.sgx.quote import get_quote


class Certificate:
    """Certificate class."""

    def __init__(
        self,
        subject_alternative_name: str,
        subject: x509.Name,
        root_path: Path,
        expiration_date: datetime,
        ratls: Optional[bytes],
    ):
        """Init constructor of SGXCertificate."""
        self.cert_path: Path = root_path / "cert.ratls.pem"
        self.key_path: Path = root_path / "key.ratls.pem"
        self.sk: ec.EllipticCurvePrivateKey = (
            ec.generate_private_key(curve=ec.SECP256R1())
            if not self.key_path.exists()
            else cast(
                ec.EllipticCurvePrivateKey,
                load_pem_private_key(data=self.key_path.read_bytes(), password=None),
            )
        )
        self.expiration_date: datetime = expiration_date
        self.cert: x509.Certificate
        self.quote: Optional[Quote] = None
        if self.key_path.exists() and self.cert_path.exists():
            self.cert = x509.load_pem_x509_certificate(data=self.cert_path.read_bytes())
            if ratls is not None:
                self.quote = get_quote_from_cert(self.cert)
        else:
            custom_extension: Optional[x509.ExtensionType] = None
            if ratls is not None:
                pubkey_hash: bytes = hashlib.sha256(
                    self.sk.public_key().public_bytes(
                        encoding=Encoding.X962,
                        format=PublicFormat.UncompressedPoint,
                    )
                ).digest()
                user_report_data: bytes = (
                    pubkey_hash + ratls if ratls is not None else pubkey_hash
                )
                self.quote = Quote.from_bytes(
                    get_quote(user_report_data=user_report_data)
                )
                custom_extension = x509.UnrecognizedExtension(
                    oid=SGX_QUOTE_EXTENSION_OID, value=bytes(self.quote)
                )
            self.cert = generate_x509(
                subject_alternative_name=subject_alternative_name,
                subject=subject,
                private_key=self.sk,
                expiration_date=self.expiration_date,
                custom_extension=custom_extension,
            )
            self.write(self.cert_path, self.key_path)

    def write(
        self, cert_path: Path, sk_path: Path, encoding: Encoding = Encoding.PEM
    ) -> None:
        """Write X509 certificate and private key to `cert_path` and `sk_path`."""
        cert_path.write_bytes(self.cert.public_bytes(encoding))
        sk_path.write_bytes(
            self.sk.private_bytes(
                encoding=Encoding.PEM,
                format=PrivateFormat.PKCS8,
                encryption_algorithm=NoEncryption(),
            )
        )


def generate_x509(
    subject_alternative_name: str,
    subject: x509.Name,
    private_key: ec.EllipticCurvePrivateKey,
    expiration_date: datetime,
    custom_extension: Optional[x509.ExtensionType] = None,
) -> x509.Certificate:
    """X509 certificate generation."""
    issuer: x509.Name = subject  # issuer=subject for self-signed certificate

    san: Union[x509.IPAddress, x509.DNSName]
    try:
        san = x509.IPAddress(ipaddress.ip_address(subject_alternative_name))
    except ValueError:
        san = x509.DNSName(subject_alternative_name)

    builder: x509.CertificateBuilder = x509.CertificateBuilder()

    builder = (
        builder.subject_name(subject)
        .issuer_name(issuer)
        .public_key(private_key.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.utcnow())
        .not_valid_after(expiration_date)
        .add_extension(
            x509.SubjectAlternativeName([san]),
            critical=False,
        )
    )

    if custom_extension is not None:
        builder = builder.add_extension(custom_extension, critical=False)

    builder = builder.add_extension(
        x509.BasicConstraints(ca=False, path_length=None),
        critical=True,
    )

    return builder.sign(private_key=private_key, algorithm=hashes.SHA256())


def to_wildcard_domain(domain: str) -> str:
    """Add wildcard to first subdomain."""
    try:
        _ = ipaddress.ip_address(domain)
    except ValueError:
        if "." not in domain:
            return domain

        subdomains: List[str] = urlparse(f"//{domain}").netloc.split(".")

        if len(subdomains) <= 2:
            return domain

        return f"*.{'.'.join(subdomains[1:])}"

    return domain



"""mse_lib_sgx.global module."""

import os
import threading
from pathlib import Path
from typing import Optional
from uuid import UUID

from cryptography import x509
from cryptography.x509.oid import NameOID

CODE_SECRET_KEY: Optional[bytes] = None

EXIT_EVENT: threading.Event = threading.Event()

ID: Optional[UUID] = None

SSL_PRIVATE_KEY: Optional[str] = None
NEED_SSL_PRIVATE_KEY: bool = False
PLAINCODE: bool = False

HOME_DIR_PATH: Path = Path(os.getenv("HOME", "/root"))
KEY_DIR_PATH: Path = Path(os.getenv("KEY_PATH", "/key"))
SECRETS_PATH: Path = Path(os.getenv("SECRETS_PATH", "/root/.cache/mse/secrets.json"))
SEALED_SECRETS_PATH: Path = Path(
    os.getenv("SEALED_SECRETS_PATH", "/root/.cache/mse/sealed_secrets.json")
)
MODULE_DIR_PATH: Path = Path(os.getenv("MODULE_PATH", "/mse-app"))

CODE_KEY_PATH: Path = KEY_DIR_PATH / "code.key"
ENCLAVE_SK_PATH: Path = KEY_DIR_PATH / "enclave.key"

SUBJECT: x509.Name = x509.Name(
    [
        x509.NameAttribute(NameOID.COUNTRY_NAME, "FR"),
        x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, "Ile-de-France"),
        x509.NameAttribute(NameOID.LOCALITY_NAME, "Paris"),
        x509.NameAttribute(NameOID.ORGANIZATION_NAME, "Cosmian Tech"),
        x509.NameAttribute(NameOID.COMMON_NAME, "localhost"),
    ]
)

TIMEOUT: Optional[int] = None



from mse_lib_sgx.certificate import to_wildcard_domain


def test_wildcard():
    assert to_wildcard_domain("localhost") == "localhost"
    assert to_wildcard_domain("127.0.0.1") == "127.0.0.1"
    assert to_wildcard_domain("cosmian.app") == "cosmian.app"
    assert to_wildcard_domain(".cosmian.app") == "*.cosmian.app"
    assert (
        to_wildcard_domain("9a5029d5-f769-4749-804e-50f6711bd509.cosmian.app")
        == "*.cosmian.app"
    )
    assert (
        to_wildcard_domain("9a5029d5-f769-4749-804e-50f6711bd509.dev.cosmian.app")
        == "*.dev.cosmian.app"
    )
    assert (
        to_wildcard_domain("9a5029d5-f769-4749-804e-50f6711bd509.staging.cosmian.app")
        == "*.staging.cosmian.app"
    )



"""mse_lib_sgx.cli.bootstrap module."""

import argparse
import asyncio
import importlib
import logging
import shutil
import sys
import sysconfig
import uuid
from datetime import datetime, timedelta
from enum import Enum
from pathlib import Path

from cryptography import x509
from hypercorn.asyncio import serve
from hypercorn.config import Config
from mse_lib_crypto.x25519 import x25519_pk_from_sk
from mse_lib_crypto.xsalsa20_poly1305 import decrypt_directory

from mse_lib_sgx import __version__, globs
from mse_lib_sgx.certificate import Certificate
from mse_lib_sgx.error import SecurityError
from mse_lib_sgx.http_server import serve as serve_sgx_secrets
from mse_lib_sgx.sgx.key import get_mrenclave_key


def parse_args() -> argparse.Namespace:
    """Argument parser."""
    parser = argparse.ArgumentParser(
        description="Bootstrap ASGI/WSGI Python web application for Gramine"
    )
    parser.add_argument(
        "application",
        type=str,
        help="ASGI application path (as module:app)",
    )

    parser.add_argument(
        "--host",
        type=str,
        default="0.0.0.0",
        help="hostname of the server",
    )
    parser.add_argument("--port", type=int, default=443, help="port of the server")
    parser.add_argument(
        "--subject",
        type=str,
        help="Subject as RFC 4514 string for the RA-TLS certificate",
    )
    parser.add_argument(
        "--san", type=str, help="Subject Alternative Name in the RA-TLS certificate"
    )
    parser.add_argument(
        "--app-dir",
        required=True,
        type=Path,
        help="path of the python web application",
    )
    parser.add_argument(
        "--id",
        required=True,
        type=uuid.UUID,
        help="identifier of the application as UUID in RFC 4122",
    )
    parser.add_argument(
        "--plaincode", action="store_true", help="unencrypted python web application"
    )
    parser.add_argument(
        "--timeout", type=int, help="seconds before closing the configuration server"
    )
    parser.add_argument(
        "--version", action="version", version=f"%(prog)s {__version__}"
    )
    parser.add_argument(
        "--debug", action="store_true", help="debug mode with more logging"
    )

    group = parser.add_mutually_exclusive_group(required=True)

    group.add_argument(
        "--ratls",
        type=int,
        metavar="EXPIRATION_DATE",
        help="generate a self-signed certificate for RA-TLS with a "
        "specific expiration date (Unix time)",
    )

    group.add_argument("--no-ssl", action="store_true", help="use HTTP without SSL")

    group.add_argument(
        "--certificate",
        type=Path,
        metavar="CERTIFICATE_PATH",
        help="custom certificate used for the SSL connection, "
        "private key must be sent through the configuration server",
    )

    return parser.parse_args()


class SslAppMode(Enum):
    """SSL Application Mode."""

    RATLS_CERTIFICATE = 1  # self-signed SGX certificate with quote
    CUSTOM_CERTIFICATE = 2  # provided by the code provider
    NO_SSL = 3  # no SSL, will be done by the SSL proxy


# pylint: disable=too-many-statements,too-many-branches
def run() -> None:
    """Entrypoint of the CLI.

    The program creates a self-signed certificate.

    Then starts a configuration server using HTTPS and this previous cert
    in order to allow the user to send some secrets params.

    Once all the secrets has been sent, three options:
    - (--self-signed) If the app owner relies on the enclave certificate,
      then start the app server using this same certificate
    - (--certificate) Start the app server using the certificate
      provided by the app owner. In that case, the certificate
      is already present in the workspace of the program
      but the private key is sent by the app owner
      when the configuration server is up.
    - (--no-ssl) If the app owner and the users trust the operator (cosmian)
      then don't use https connection.
    """
    args: argparse.Namespace = parse_args()

    globs.HOME_DIR_PATH.mkdir(exist_ok=True)
    globs.KEY_DIR_PATH.mkdir(exist_ok=True)

    logging.basicConfig(
        level=logging.DEBUG if args.debug else logging.INFO,
        format="[%(asctime)s] [%(levelname)s] %(message)s",
    )

    if args.plaincode:
        globs.PLAINCODE = True

    if args.timeout:
        globs.TIMEOUT = args.timeout

    ssl_private_key_path = None
    expiration_date = datetime.now() + timedelta(hours=10)

    ssl_app_mode: SslAppMode
    if args.no_ssl:
        # The conf server use the self-signed cert
        # No ssl for the app server
        ssl_app_mode = SslAppMode.NO_SSL
    elif args.certificate:
        # The conf server use the self-signed cert
        # The app server use the app owner cert
        ssl_app_mode = SslAppMode.CUSTOM_CERTIFICATE
        ssl_private_key_path = globs.KEY_DIR_PATH / "key.pem"
    else:
        # The conf server and the app server will use the same self-signed cert
        ssl_app_mode = SslAppMode.RATLS_CERTIFICATE
        expiration_date = datetime.utcfromtimestamp(args.ratls)

    logging.info("Generating self-signed certificate...")

    if not globs.ENCLAVE_SK_PATH.exists():
        globs.ENCLAVE_SK_PATH.write_bytes(get_mrenclave_key())

    enclave_pk: bytes = x25519_pk_from_sk(globs.ENCLAVE_SK_PATH.read_bytes())

    if len(enclave_pk) != 32:
        raise SecurityError("Bad enclave pk length!")

    cert: Certificate = Certificate(
        subject_alternative_name=args.san if args.san else "localhost",
        subject=(
            x509.Name.from_rfc4514_string(args.subject)
            if args.subject
            else globs.SUBJECT
        ),
        root_path=globs.KEY_DIR_PATH,
        expiration_date=expiration_date,
        ratls=enclave_pk,
    )

    if not globs.MODULE_DIR_PATH.exists():
        logging.info("Starting the configuration server...")
        # The app owner will send:
        # - the uuid of the app (see as an uniq token allowing to query the API)
        # - the key to decrypt the code
        # - (optional) the SSL private key if AppConnection.OWNER_CERTFICIATE
        serve_sgx_secrets(
            hostname=args.host,
            port=args.port,
            certificate=cert,
            app_id=args.id,
            need_ssl_private_key=ssl_app_mode == SslAppMode.CUSTOM_CERTIFICATE,
            timeout=globs.TIMEOUT,
        )

        if not globs.PLAINCODE and globs.CODE_SECRET_KEY is None:
            raise SecurityError("Code secret key not provided")

        if not globs.PLAINCODE and globs.CODE_SECRET_KEY is not None:
            globs.CODE_KEY_PATH.write_bytes(globs.CODE_SECRET_KEY)

        if (
            ssl_app_mode == SslAppMode.CUSTOM_CERTIFICATE
            and globs.SSL_PRIVATE_KEY
            and ssl_private_key_path is not None
        ):
            ssl_private_key_path.write_text(globs.SSL_PRIVATE_KEY)

        if globs.PLAINCODE:
            shutil.copytree(
                src=args.app_dir, dst=globs.MODULE_DIR_PATH, copy_function=shutil.copy
            )
        else:
            globs.MODULE_DIR_PATH.mkdir()
            decrypt_directory(
                dir_path=args.app_dir,
                key=globs.CODE_KEY_PATH.read_bytes(),
                ext=".enc",
                out_dir_path=globs.MODULE_DIR_PATH,
            )

    config_map = {
        "bind": f"{args.host}:{args.port}",
        "alpn_protocols": ["h2"],
        "workers": 1,
        "accesslog": "-",
        "errorlog": "-",
        "worker_class": "uvloop",
        "wsgi_max_body_size": 2 * 1024 * 1024 * 1024,  # 2 GB
    }

    if ssl_app_mode == SslAppMode.CUSTOM_CERTIFICATE:
        config_map["certfile"] = args.certificate
        config_map["keyfile"] = ssl_private_key_path
    elif ssl_app_mode == SslAppMode.RATLS_CERTIFICATE:
        config_map["certfile"] = cert.cert_path
        config_map["keyfile"] = cert.key_path

    config = Config.from_mapping(config_map)

    logging.info("Loading the application...")
    module_name, application_name = args.application.split(":")

    sys.path.append(f"{globs.MODULE_DIR_PATH}")

    logging.debug("MODULE_PATH=%s", globs.MODULE_DIR_PATH)
    logging.debug("sys.path: %s", sys.path)
    logging.debug("sysconfig.get_paths(): %s", sysconfig.get_paths())
    logging.debug("application: %s", args.application)

    application = getattr(importlib.import_module(module_name), application_name)

    logging.info("Starting the application (mode=%s)...", ssl_app_mode.name)
    asyncio.run(serve(application, config))



"""mse_lib_sgx.sgx.quote module."""

from mse_lib_sgx.error import SGXError

SGX_QUOTE_MAX_SIZE: int = 8192


def get_quote(user_report_data: bytes) -> bytes:
    """Read the quote if inside Intel SGX enclave."""
    if len(user_report_data) > 64:
        raise SGXError("user_report_data must be at most 64 bytes")

    attestation_type: str
    try:
        with open("/dev/attestation/attestation_type", "rb") as f:
            attestation_type = f.read(32).decode("utf-8").strip()

        if attestation_type != "dcap":
            raise SGXError(f"Only DCAP supported, found '{attestation_type}'")

        with open("/dev/attestation/user_report_data", "wb") as f:
            f.write(user_report_data)

        with open("/dev/attestation/quote", "rb") as f:
            return f.read(SGX_QUOTE_MAX_SIZE)
    except FileNotFoundError as exc:
        raise SGXError("Not running inside Intel SGX") from exc
